import urllib.request

import pytest

from woodwork.demo import load_retail
from woodwork.logical_types import (
    Boolean,
    Categorical,
    Datetime,
    Double,
    Integer,
    NaturalLanguage,
)
from woodwork.table_schema import TableSchema


@pytest.fixture(autouse=True)
def set_testing_headers():
    opener = urllib.request.build_opener()
    opener.addheaders = [("Testing", "True")]
    urllib.request.install_opener(opener)


def test_load_retail_diff():
    nrows = 10
    df = load_retail(nrows=nrows, init_woodwork=False)
    assert df.ww.schema is None
    assert df.shape[0] == nrows
    nrows_second = 11
    df = load_retail(nrows=nrows_second, init_woodwork=False)
    assert df.ww.schema is None
    assert df.shape[0] == nrows_second

    assert "order_product_id" in df.columns
    assert df["order_product_id"].is_unique


def test_load_retail():
    df = load_retail(nrows=10, init_woodwork=True)
    assert isinstance(df.ww.schema, TableSchema)

    expected_logical_types = {
        "order_product_id": Categorical,
        "order_id": Categorical,
        "product_id": Categorical,
        "description": NaturalLanguage,
        "quantity": Integer,
        "order_date": Datetime,
        "unit_price": Double,
        "customer_name": Categorical,
        "country": Categorical,
        "total": Double,
        "cancelled": Boolean,
    }

    for col_name, column in df.ww.columns.items():
        assert isinstance(column.logical_type, expected_logical_types[col_name])

    assert df.ww.index == "order_product_id"
    assert df.ww.time_index == "order_date"
    assert df.ww.semantic_tags["order_product_id"] == {"index"}
    assert df.ww.semantic_tags["order_date"] == {"time_index"}



import re
from unittest.mock import patch

import pytest

from woodwork.exceptions import ColumnNotPresentError
from woodwork.logical_types import (
    Categorical,
    Datetime,
    Double,
    Integer,
    NaturalLanguage,
    Ordinal,
)
from woodwork.table_schema import (
    TableSchema,
    _check_column_descriptions,
    _check_column_metadata,
    _check_column_names,
    _check_column_origins,
    _check_index,
    _check_logical_types,
    _check_semantic_tags,
    _check_table_metadata,
    _check_time_index,
    _check_use_standard_tags,
    _validate_params,
)


def test_validate_params_errors(sample_column_names):
    error_message = "Table name must be a string"
    with pytest.raises(TypeError, match=error_message):
        _validate_params(
            column_names=sample_column_names,
            name=1,
            index=None,
            time_index=None,
            logical_types=None,
            table_metadata=None,
            column_metadata=None,
            semantic_tags=None,
            column_descriptions=None,
            column_origins=None,
            use_standard_tags=False,
        )


def test_check_index_errors(sample_column_names):
    error_message = "Specified index column `foo` not found in TableSchema."
    with pytest.raises(ColumnNotPresentError, match=error_message):
        _check_index(column_names=sample_column_names, index="foo")


def test_check_time_index_errors(sample_column_names):
    error_message = "Specified time index column `foo` not found in TableSchema"
    with pytest.raises(ColumnNotPresentError, match=error_message):
        _check_time_index(
            column_names=sample_column_names,
            time_index="foo",
            logical_type=Integer,
        )

    error_msg = "Time index column must be a Datetime or numeric column"
    with pytest.raises(TypeError, match=error_msg):
        _check_time_index(
            column_names=sample_column_names,
            time_index="full_name",
            logical_type=NaturalLanguage,
        )


def test_check_column_names(sample_column_names):
    error_message = "Column names must be a list or set"
    with pytest.raises(TypeError, match=error_message):
        _check_column_names(column_names=int)

    sample_column_names.append("id")
    with pytest.raises(
        IndexError,
        match="TableSchema cannot contain duplicate columns names",
    ):
        _check_column_names(sample_column_names)


def test_check_logical_types_errors(sample_column_names):
    error_message = "logical_types must be a dictionary"
    with pytest.raises(TypeError, match=error_message):
        _check_logical_types(sample_column_names, logical_types="type")

    bad_logical_types_keys = {
        "full_name": None,
        "age": None,
        "birthday": None,
        "occupation": None,
    }
    error_message = re.escape(
        "logical_types contains columns that are not present in TableSchema: ['birthday', 'occupation']",
    )
    with pytest.raises(ColumnNotPresentError, match=error_message):
        _check_logical_types(sample_column_names, bad_logical_types_keys)

    bad_logical_types_keys = {
        "id": None,
        "full_name": None,
        "email": None,
        "phone_number": None,
        "age": None,
    }
    error_message = re.escape(
        "logical_types is missing columns that are present in TableSchema: "
        "['boolean', 'categorical', 'datetime_with_NaT', 'double', 'double_with_nan', "
        "'integer', 'ip_address', 'is_registered', 'nullable_integer', 'signup_date', 'url']",
    )
    with pytest.raises(ColumnNotPresentError, match=error_message):
        _check_logical_types(sample_column_names, bad_logical_types_keys)

    bad_logical_types_keys = {"email": 1}
    error_message = (
        "Logical Types must be of the LogicalType class "
        "and registered in Woodwork's type system. "
        "1 does not meet that criteria."
    )

    with pytest.raises(TypeError, match=error_message):
        _check_logical_types(
            sample_column_names,
            bad_logical_types_keys,
            require_all_cols=False,
        )

    bad_logical_types_keys = {
        "email": "NaturalLanguage",
    }
    error_message = (
        "Logical Types must be of the LogicalType class "
        "and registered in Woodwork's type system. "
        "NaturalLanguage does not meet that criteria."
    )

    with pytest.raises(TypeError, match=error_message):
        _check_logical_types(
            sample_column_names,
            bad_logical_types_keys,
            require_all_cols=False,
        )


def test_check_semantic_tags_errors(sample_column_names):
    error_message = "semantic_tags must be a dictionary"
    with pytest.raises(TypeError, match=error_message):
        _check_semantic_tags(sample_column_names, semantic_tags="type")

    bad_semantic_tags_keys = {
        "full_name": None,
        "age": None,
        "birthday": None,
        "occupation": None,
    }
    error_message = re.escape(
        "semantic_tags contains columns that are not present in TableSchema: ['birthday', 'occupation']",
    )
    with pytest.raises(ColumnNotPresentError, match=error_message):
        _check_semantic_tags(sample_column_names, bad_semantic_tags_keys)

    error_message = "semantic_tags for id must be a string, set or list"
    with pytest.raises(TypeError, match=error_message):
        _check_semantic_tags(sample_column_names, {"id": 1})


def test_check_table_metadata_errors():
    error_message = "Table metadata must be a dictionary."
    with pytest.raises(TypeError, match=error_message):
        _check_table_metadata("test")


def test_check_column_metadata_errors(sample_column_names):
    error_message = "Column metadata must be a dictionary."
    with pytest.raises(TypeError, match=error_message):
        _check_column_metadata(sample_column_names, column_metadata="test")

    column_metadata = {"invalid_col": {"description": "not a valid column"}}
    err_msg = re.escape(
        "column_metadata contains columns that are not present in TableSchema: ['invalid_col']",
    )
    with pytest.raises(ColumnNotPresentError, match=err_msg):
        _check_column_metadata(sample_column_names, column_metadata=column_metadata)


def test_check_column_description_errors(sample_column_names):
    error_message = "column_descriptions must be a dictionary"
    with pytest.raises(TypeError, match=error_message):
        _check_column_descriptions(sample_column_names, column_descriptions="test")

    column_descriptions = {"invalid_col": "a description"}
    err_msg = re.escape(
        "column_descriptions contains columns that are not present in TableSchema: ['invalid_col']",
    )
    with pytest.raises(ColumnNotPresentError, match=err_msg):
        _check_column_descriptions(
            sample_column_names,
            column_descriptions=column_descriptions,
        )


def test_check_column_origin_errors(sample_column_names):
    error_message = "column_origins must be a dictionary or a string"
    with pytest.raises(TypeError, match=error_message):
        _check_column_origins(sample_column_names, column_origins=123)

    column_origins = {"invalid_col": "base"}
    err_msg = re.escape(
        "column_origins contains columns that are not present in TableSchema: ['invalid_col']",
    )
    with pytest.raises(ColumnNotPresentError, match=err_msg):
        _check_column_origins(sample_column_names, column_origins=column_origins)


def test_check_use_standard_tags_errors(sample_column_names):
    error_message = "use_standard_tags must be a dictionary or a boolean"
    with pytest.raises(TypeError, match=error_message):
        _check_use_standard_tags(sample_column_names, use_standard_tags=1)

    error_message = re.escape(
        "use_standard_tags contains columns that are not present in TableSchema: ['invalid_col']",
    )
    with pytest.raises(ColumnNotPresentError, match=error_message):
        _check_use_standard_tags(
            sample_column_names,
            use_standard_tags={"invalid_col": True},
        )

    error_message = "use_standard_tags for column id must be a boolean"
    with pytest.raises(TypeError, match=error_message):
        _check_use_standard_tags(sample_column_names, use_standard_tags={"id": 1})


def test_schema_init(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(sample_column_names, sample_inferred_logical_types)

    assert schema.name is None
    assert schema.index is None
    assert schema.time_index is None

    assert set(schema.columns.keys()) == set(sample_column_names)


def test_schema_init_with_name(sample_column_names, sample_inferred_logical_types):
    schema = TableSchema(
        sample_column_names,
        sample_inferred_logical_types,
        name="schema",
    )

    assert schema.name == "schema"
    assert schema.index is None
    assert schema.time_index is None


def test_schema_init_with_name_and_indices(
    sample_column_names,
    sample_inferred_logical_types,
):
    schema = TableSchema(
        sample_column_names,
        sample_inferred_logical_types,
        name="schema",
        index="id",
        time_index="signup_date",
    )

    assert schema.name == "schema"
    assert schema.index == "id"
    assert schema.time_index == "signup_date"
    assert isinstance(schema.columns[schema.time_index].logical_type, Datetime)


def test_schema_with_numeric_time_index(
    sample_column_names,
    sample_inferred_logical_types,
):
    # Set a numeric time index on init
    schema = TableSchema(
        sample_column_names,
        logical_types={**sample_inferred_logical_types, **{"signup_date": Integer}},
        time_index="signup_date",
        use_standard_tags=True,
    )
    date_col = schema.columns["signup_date"]
    assert schema.time_index == "signup_date"
    assert isinstance(date_col.logical_type, Integer)
    assert date_col.semantic_tags == {"time_index", "numeric"}

    # Specify logical type for time index on init
    schema = TableSchema(
        sample_column_names,
        logical_types={**sample_inferred_logical_types, **{"signup_date": Double}},
        time_index="signup_date",
        use_standard_tags=True,
    )
    date_col = schema.columns["signup_date"]
    assert schema.time_index == "signup_date"
    assert isinstance(date_col.logical_type, Double)
    assert date_col.semantic_tags == {"time_index", "numeric"}


def test_schema_init_with_logical_type_classes(
    sample_column_names,
    sample_correct_logical_types,
):
    schema = TableSchema(
        sample_column_names,
        logical_types=sample_correct_logical_types,
        name="schema",
    )
    instantiated_ltypes = {
        name: ltype() for name, ltype in sample_correct_logical_types.items()
    }
    assert schema.logical_types == instantiated_ltypes


def test_raises_error_setting_index_tag_directly(
    sample_column_names,
    sample_inferred_logical_types,
):
    error_msg = re.escape(
        "Cannot add 'index' tag directly for column id. To set a column as the index, "
        "use DataFrame.ww.set_index() instead.",
    )
    with pytest.raises(ValueError, match=error_msg):
        semantic_tags = {"id": "index"}
        TableSchema(
            sample_column_names,
            sample_inferred_logical_types,
            name="schema",
            semantic_tags=semantic_tags,
            use_standard_tags=False,
        )


def test_raises_error_setting_time_index_tag_directly(
    sample_column_names,
    sample_inferred_logical_types,
):
    error_msg = re.escape(
        "Cannot add 'time_index' tag directly for column signup_date. To set a column as the time index, "
        "use DataFrame.ww.set_time_index() instead.",
    )
    with pytest.raises(ValueError, match=error_msg):
        semantic_tags = {"signup_date": "time_index"}
        TableSchema(
            sample_column_names,
            sample_inferred_logical_types,
            name="schema",
            semantic_tags=semantic_tags,
            use_standard_tags=False,
        )


def test_schema_init_with_semantic_tags(
    sample_column_names,
    sample_inferred_logical_types,
):
    semantic_tags = {"id": "custom_tag"}
    schema = TableSchema(
        sample_column_names,
        sample_inferred_logical_types,
        name="schema",
        semantic_tags=semantic_tags,
        use_standard_tags=False,
    )

    id_semantic_tags = schema.columns["id"].semantic_tags
    assert isinstance(id_semantic_tags, set)
    assert len(id_semantic_tags) == 1
    assert "custom_tag" in id_semantic_tags


def test_schema_adds_standard_semantic_tags(
    sample_column_names,
    sample_inferred_logical_types,
):
    schema = TableSchema(
        sample_column_names,
        logical_types={**sample_inferred_logical_types, **{"id": Categorical}},
        use_standard_tags=True,
        name="schema",
    )

    assert schema.semantic_tags["id"] == {"category"}
    assert schema.semantic_tags["age"] == {"numeric"}

    schema = TableSchema(
        sample_column_names,
        logical_types={**sample_inferred_logical_types, **{"id": Categorical}},
        name="schema",
        use_standard_tags=False,
    )

    assert schema.semantic_tags["id"] == set()
    assert schema.semantic_tags["age"] == set()


def test_semantic_tags_during_init(sample_column_names, sample_inferred_logical_types):
    semantic_tags = {
        "full_name": "tag1",
        "email": ["tag2"],
        "phone_number": ["tag3"],
        "signup_date": ["secondary_time_index"],
        "age": ["numeric", "age"],
    }
    expected_types = {
        "full_name": {"tag1"},
        "email": {"tag2"},
        "phone_number": {"tag3"},
        "signup_date": {"secondary_time_index"},
        "age": {"numeric", "age"},
    }
    schema = TableSchema(
        sample_column_names,
        sample_inferred_logical_types,
        semantic_tags=semantic_tags,
    )
    assert schema.columns["full_name"].semantic_tags == expected_types["full_name"]
    assert schema.columns["email"].semantic_tags == expected_types["email"]
    assert (
        schema.columns["phone_number"].semantic_tags == expected_types["phone_number"]
    )
    assert schema.columns["signup_date"].semantic_tags == expected_types["signup_date"]
    assert schema.columns["age"].semantic_tags == expected_types["age"]


def test_semantic_tag_errors(sample_column_names, sample_inferred_logical_types):
    error_message = "semantic_tags for id must be a string, set or list"
    with pytest.raises(TypeError, match=error_message):
        TableSchema(
            sample_column_names,
            sample_inferred_logical_types,
            semantic_tags={"id": int},
        )

    error_message = "semantic_tags for id must be a string, set or list"
    with pytest.raises(TypeError, match=error_message):
        TableSchema(
            sample_column_names,
            sample_inferred_logical_types,
            semantic_tags={"id": {"index": {}, "time_index": {}}},
        )

    error_message = "semantic_tags for id must contain only strings"
    with pytest.raises(TypeError, match=error_message):
        TableSchema(
            sample_column_names,
            sample_inferred_logical_types,
            semantic_tags={"id": ["index", 1]},
        )


def test_index_replacing_standard_tags(
    sample_column_names,
    sample_inferred_logical_types,
):
    schema = TableSchema(
        sample_column_names,
        sample_inferred_logical_types,
        use_standard_tags=True,
    )
    assert schema.columns["id"].semantic_tags == {"numeric"}

    schema = TableSchema(sample_column_names, sample_inferred_logical_types, index="id")
    assert schema.columns["id"].semantic_tags == {"index"}


def test_schema_init_with_col_descriptions(
    sample_column_names,
    sample_inferred_logical_types,
):
    descriptions = {"age": "age of the user", "signup_date": "date of account creation"}
    schema = TableSchema(
        sample_column_names,
        sample_inferred_logical_types,
        column_descriptions=descriptions,
    )
    for name, column in schema.columns.items():
        assert column.description == descriptions.get(name)


def test_schema_col_descriptions_errors(
    sample_column_names,
    sample_inferred_logical_types,
):
    err_msg = "column_descriptions must be a dictionary"
    with pytest.raises(TypeError, match=err_msg):
        TableSchema(
            sample_column_names,
            sample_inferred_logical_types,
            column_descriptions=34,
        )

    descriptions = {
        "invalid_col": "not a valid column",
        "signup_date": "date of account creation",
    }
    err_msg = re.escape(
        "column_descriptions contains columns that are not present in TableSchema: ['invalid_col']",
    )
    with pytest.raises(ColumnNotPresentError, match=err_msg):
        TableSchema(
            sample_column_names,
            sample_inferred_logical_types,
            column_descriptions=descriptions,
        )


def test_schema_init_with_col_origins(
    sample_column_names,
    sample_inferred_logical_types,
):
    origins = {"age": "base", "signup_date": "engineered"}
    schema = TableSchema(
        sample_column_names,
        sample_inferred_logical_types,
        column_origins=origins,
    )
    for name, column in schema.columns.items():
        assert column.origin == origins.get(name)

    schema_single_origin = TableSchema(
        sample_column_names,
        sample_inferred_logical_types,
        column_origins="base",
    )
    for name, column in schema_single_origin.columns.items():
        assert column.origin == "base"


def test_schema_col_origins_errors(sample_column_names, sample_inferred_logical_types):
    err_msg = "column_origins must be a dictionary or a string"
    with pytest.raises(TypeError, match=err_msg):
        TableSchema(
            sample_column_names,
            sample_inferred_logical_types,
            column_origins=34,
        )

    origins = {"invalid_col": "not a valid column", "signup_date": "base"}
    err_msg = re.escape(
        "column_origins contains columns that are not present in TableSchema: ['invalid_col']",
    )
    with pytest.raises(ColumnNotPresentError, match=err_msg):
        TableSchema(
            sample_column_names,
            sample_inferred_logical_types,
            column_origins=origins,
        )

    origins = {"signup_date": 1}
    err_msg = "Column origin must be a string"
    with pytest.raises(TypeError, match=err_msg):
        TableSchema(
            sample_column_names,
            sample_inferred_logical_types,
            column_origins=origins,
        )


def test_schema_init_with_column_metadata(
    sample_column_names,
    sample_inferred_logical_types,
):
    column_metadata = {
        "age": {"interesting_values": [33]},
        "signup_date": {"description": "date of account creation"},
    }
    schema = TableSchema(
        sample_column_names,
        sample_inferred_logical_types,
        column_metadata=column_metadata,
    )
    for name, column in schema.columns.items():
        assert column.metadata == (column_metadata.get(name) or {})


@patch("woodwork.table_schema._validate_not_setting_index_tags")
@patch("woodwork.table_schema._check_time_index")
@patch("woodwork.table_schema._check_index")
@patch("woodwork.table_schema._validate_params")
def test_validation_methods_called(
    mock_validate_params,
    mock_check_index,
    mock_check_time_index,
    mock_validate_not_setting_index,
    sample_column_names,
    sample_inferred_logical_types,
):
    assert not mock_validate_params.called
    assert not mock_check_index.called
    assert not mock_check_time_index.called
    assert not mock_validate_not_setting_index.called

    not_validated_schema = TableSchema(
        sample_column_names,
        sample_inferred_logical_types,
        index="id",
        time_index="signup_date",
        validate=False,
    )
    assert not mock_validate_params.called
    assert not mock_check_index.called
    assert not mock_check_time_index.called
    assert not mock_validate_not_setting_index.called

    validated_schema = TableSchema(
        sample_column_names,
        sample_inferred_logical_types,
        index="id",
        time_index="signup_date",
        validate=True,
    )
    assert mock_validate_params.called
    assert mock_check_index.called
    assert mock_check_time_index.called
    assert mock_validate_not_setting_index.called

    assert validated_schema == not_validated_schema


def test_use_standard_tags_from_bool(
    sample_column_names,
    sample_inferred_logical_types,
):
    standard_tags_schema = TableSchema(
        sample_column_names,
        sample_inferred_logical_types,
        use_standard_tags=True,
    )
    assert set(standard_tags_schema.columns.keys()) == set(
        standard_tags_schema.use_standard_tags.keys(),
    )
    assert all([*standard_tags_schema.use_standard_tags.values()])
    assert standard_tags_schema.semantic_tags["id"] == {"numeric"}

    no_standard_tags_schema = TableSchema(
        sample_column_names,
        sample_inferred_logical_types,
        use_standard_tags=False,
    )
    assert set(no_standard_tags_schema.columns.keys()) == set(
        no_standard_tags_schema.use_standard_tags.keys(),
    )
    assert not any([*no_standard_tags_schema.use_standard_tags.values()])
    assert no_standard_tags_schema.semantic_tags["id"] == set()


def test_use_standard_tags_from_dict(
    sample_column_names,
    sample_inferred_logical_types,
):
    default_schema = TableSchema(
        sample_column_names,
        sample_inferred_logical_types,
        use_standard_tags={col_name: False for col_name in sample_column_names},
    )
    assert default_schema.use_standard_tags == {
        col_name: False for col_name in sample_column_names
    }

    use_standard_tags = {
        "id": True,
        "full_name": False,
        "email": True,
        "phone_number": True,
        "age": False,
        "signup_date": True,
        "is_registered": False,
        "double": False,
        "double_with_nan": False,
        "integer": False,
        "nullable_integer": False,
        "boolean": False,
        "categorical": False,
        "datetime_with_NaT": False,
        "url": False,
        "ip_address": False,
    }
    full_dict_schema = TableSchema(
        sample_column_names,
        sample_inferred_logical_types,
        use_standard_tags=use_standard_tags,
    )
    assert full_dict_schema.use_standard_tags == use_standard_tags

    partial_dict_schema = TableSchema(
        sample_column_names,
        sample_inferred_logical_types,
        use_standard_tags={
            "id": True,
            "email": True,
            "phone_number": True,
            "signup_date": True,
        },
    )
    assert full_dict_schema.use_standard_tags == partial_dict_schema.use_standard_tags
    assert full_dict_schema == partial_dict_schema

    partial_dict_default_schema = TableSchema(
        sample_column_names,
        sample_inferred_logical_types,
        use_standard_tags={
            "id": False,
            "email": False,
            "phone_number": False,
            "signup_date": False,
        },
    )
    assert (
        default_schema.use_standard_tags
        == partial_dict_default_schema.use_standard_tags
    )
    assert default_schema == partial_dict_default_schema


def test_ordinal_without_init():
    schema = TableSchema(
        column_names=["ordinal_col"],
        logical_types={"ordinal_col": Ordinal},
    )
    assert isinstance(schema.logical_types["ordinal_col"], Ordinal)
    assert schema.logical_types["ordinal_col"].order is None



from unittest.mock import patch

import woodwork as ww
from woodwork.accessor_utils import _is_dask_series, _is_spark_series
from woodwork.logical_types import (
    URL,
    Boolean,
    BooleanNullable,
    Categorical,
    Datetime,
    Double,
    EmailAddress,
    Integer,
    IntegerNullable,
    IPAddress,
    LogicalType,
    NaturalLanguage,
    PhoneNumber,
    PostalCode,
    Timedelta,
    Unknown,
)
from woodwork.tests.testing_utils import to_pandas
from woodwork.type_sys.type_system import (
    DEFAULT_INFERENCE_FUNCTIONS,
    DEFAULT_RELATIONSHIPS,
    DEFAULT_TYPE,
    TypeSystem,
)
from woodwork.utils import import_or_none

UNSUPPORTED_SPARK_DTYPES = [
    "int32",
    "intp",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
    "uintp",
    "float_",
    "object",
    "category",
]

ps = import_or_none("pyspark.pandas")


def get_spark_dtypes(dtypes):
    return [dtype for dtype in dtypes if dtype not in UNSUPPORTED_SPARK_DTYPES]


def test_integer_inference(integers):
    dtypes = ["int8", "int16", "int32", "int64", "intp", "int", "Int64"]
    if _is_spark_series(integers[0]):
        dtypes = get_spark_dtypes(dtypes)

    for series in integers:
        for dtype in dtypes:
            inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
            assert isinstance(inferred_type, Integer)


def test_double_inference(doubles):
    dtypes = ["float", "float32", "float64", "float_"]
    if _is_spark_series(doubles[0]):
        dtypes = get_spark_dtypes(dtypes)

    for series in doubles:
        for dtype in dtypes:
            inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
            assert isinstance(inferred_type, Double)


def test_boolean_inference(bools):
    dtypes = ["bool", "boolean"]

    for series in bools:
        for dtype in dtypes:
            if _is_dask_series(series):
                series = series.compute()
            cast_series = series
            if True in series.values:
                cast_series = series.astype(dtype)
            inferred_type = ww.type_system.infer_logical_type(cast_series)
            if to_pandas(cast_series).isnull().any():
                assert isinstance(inferred_type, BooleanNullable)
            else:
                assert isinstance(inferred_type, Boolean)


def test_datetime_inference(datetimes):
    dtypes = ["object", "string", "datetime64[ns]"]
    if _is_spark_series(datetimes[0]):
        dtypes = get_spark_dtypes(dtypes)

    for series in datetimes:
        for dtype in dtypes:
            inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
            assert isinstance(inferred_type, Datetime)


def test_email_inference(emails):
    dtypes = ["object", "string"]
    if _is_spark_series(emails[0]):
        dtypes = get_spark_dtypes(dtypes)

    for series in emails:
        for dtype in dtypes:
            inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
            assert isinstance(inferred_type, EmailAddress)


def test_email_inference_failure(bad_emails):
    dtypes = ["object", "string"]
    if _is_spark_series(bad_emails[0]):
        dtypes = get_spark_dtypes(dtypes)

    for series in bad_emails:
        if _is_spark_series(series) and isinstance(series.iloc[0], ps.series.Row):
            continue

        for dtype in dtypes:
            inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
            assert not isinstance(inferred_type, EmailAddress)


def test_categorical_inference(categories):
    dtypes = ["object", "string", "category"]
    if _is_spark_series(categories[0]):
        dtypes = get_spark_dtypes(dtypes)
    for ind, series in enumerate(categories):
        if ind == len(categories) - 1:
            dtypes = ["string", "category"]
        for dtype in dtypes:
            expected_ltype = Categorical
            if ind in [1, 3] and dtype == "object":
                expected_ltype = Integer
            inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
            assert isinstance(inferred_type, expected_ltype)


def test_postal_inference(postal):
    dtypes = ["category", "string"]
    for series in postal:
        if _is_spark_series(series):
            dtypes = get_spark_dtypes(dtypes)
        for dtype in dtypes:
            inferred_dtype = ww.type_system.infer_logical_type(series.astype(dtype))
            assert isinstance(inferred_dtype, PostalCode)


def test_natural_language_inference(natural_language):
    dtypes = ["object", "string"]
    if _is_spark_series(natural_language[0]):
        dtypes = get_spark_dtypes(dtypes)
    for series in natural_language:
        for dtype in dtypes:
            inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
            assert isinstance(inferred_type, NaturalLanguage)


@patch("woodwork.type_sys.inference_functions.natural_language_func")
def test_nl_inference_called_on_no_other_matches(nl_mock, pandas_natural_language):
    assert isinstance(
        ww.type_system.infer_logical_type(pandas_natural_language[0]),
        NaturalLanguage,
    )
    new_type_sys = TypeSystem(
        inference_functions=DEFAULT_INFERENCE_FUNCTIONS,
        relationships=DEFAULT_RELATIONSHIPS,
        default_type=DEFAULT_TYPE,
    )
    new_type_sys.inference_functions[NaturalLanguage] = nl_mock
    _ = new_type_sys.infer_logical_type(pandas_natural_language[0])
    assert nl_mock.called


@patch("woodwork.type_sys.inference_functions.natural_language_func")
def test_nl_inference_called_with_unknown_type(nl_mock, pandas_strings):
    assert isinstance(ww.type_system.infer_logical_type(pandas_strings[0]), Unknown)
    new_type_sys = TypeSystem(
        inference_functions=DEFAULT_INFERENCE_FUNCTIONS,
        relationships=DEFAULT_RELATIONSHIPS,
        default_type=DEFAULT_TYPE,
    )
    new_type_sys.inference_functions[NaturalLanguage] = nl_mock
    _ = new_type_sys.infer_logical_type(pandas_strings[0])
    assert nl_mock.called


@patch("woodwork.type_sys.inference_functions.natural_language_func")
def test_nl_inference_not_called_with_other_matches(nl_mock, pandas_integers):
    assert isinstance(ww.type_system.infer_logical_type(pandas_integers[0]), Integer)
    new_type_sys = TypeSystem(
        inference_functions=DEFAULT_INFERENCE_FUNCTIONS,
        relationships=DEFAULT_RELATIONSHIPS,
        default_type=DEFAULT_TYPE,
    )
    new_type_sys.inference_functions[NaturalLanguage] = nl_mock
    _ = new_type_sys.infer_logical_type(pandas_integers[0])
    assert not nl_mock.called


def test_categorical_inference_based_on_dtype(categories_dtype):
    """
    This test specifically targets the case in which a series can be inferred
    to be categorical strictly from its pandas dtype, but would otherwise be
    inferred as some other type.
    """
    inferred_type = ww.type_system.infer_logical_type(categories_dtype["cat"])
    assert isinstance(inferred_type, Categorical)

    inferred_type = ww.type_system.infer_logical_type(categories_dtype["non_cat"])
    assert not isinstance(inferred_type, Categorical)


def test_categorical_integers_inference(integers):
    with ww.config.with_options(numeric_categorical_threshold=0.5):
        dtypes = ["int8", "int16", "int32", "int64", "intp", "int", "Int64"]
        if _is_spark_series(integers[0]):
            dtypes = get_spark_dtypes(dtypes)
        for series in integers:
            for dtype in dtypes:
                inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
                assert isinstance(inferred_type, Categorical)


def test_categorical_double_inference(doubles):
    with ww.config.with_options(numeric_categorical_threshold=0.5):
        dtypes = ["float", "float32", "float64", "float_"]
        if _is_spark_series(doubles[0]):
            dtypes = get_spark_dtypes(dtypes)
        for series in doubles:
            for dtype in dtypes:
                inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
                assert isinstance(inferred_type, Categorical)


def test_timedelta_inference(timedeltas):
    dtypes = ["timedelta64[ns]"]
    for series in timedeltas:
        for dtype in dtypes:
            inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
            assert isinstance(inferred_type, Timedelta)


def test_unknown_inference(strings):
    dtypes = ["object", "string"]
    if _is_spark_series(strings[0]):
        dtypes = get_spark_dtypes(dtypes)

    for series in strings:
        for dtype in dtypes:
            inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
            assert isinstance(inferred_type, Unknown)


def test_unknown_inference_all_null(nulls):
    dtypes = ["object", "string", "category", "datetime64[ns]"]

    for ind, series in enumerate(nulls):
        if ind == len(nulls) - 1:
            dtypes = ["object", "string", "category"]
        if _is_spark_series(nulls[0]):
            dtypes = get_spark_dtypes(dtypes)
        for dtype in dtypes:
            inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
            inferred_type.transform(series)
            assert isinstance(inferred_type, Unknown)


def test_unknown_inference_empty_series(empty_series):
    inferred_type = ww.type_system.infer_logical_type(empty_series.astype("string"))
    assert isinstance(inferred_type, Unknown)


def test_pdna_inference(pdnas):
    expected_logical_types = [
        NaturalLanguage,
        Unknown,
        IntegerNullable,
        BooleanNullable,
    ]

    for index, series in enumerate(pdnas):
        inferred_type = ww.type_system.infer_logical_type(series)
        assert isinstance(inferred_type, expected_logical_types[index])


def test_updated_ltype_inference(integers, type_sys):
    inference_fn = type_sys.inference_functions[ww.logical_types.Integer]
    type_sys.remove_type(ww.logical_types.Integer)

    class Integer(LogicalType):
        primary_dtype = "string"

    type_sys.add_type(Integer, inference_function=inference_fn)

    dtypes = ["int8", "int16", "int32", "int64", "intp", "int", "Int64"]
    if _is_spark_series(integers[0]):
        dtypes = get_spark_dtypes(dtypes)

    for series in integers:
        for dtype in dtypes:
            inferred_type = type_sys.infer_logical_type(series.astype(dtype))
            assert isinstance(inferred_type, Integer)
            assert inferred_type.primary_dtype == "string"


def test_inference_randomly_sampled(large_df, type_sys):
    large_df.ww.init()

    inferred_type = large_df.ww.logical_types["int_nullable"]
    assert isinstance(inferred_type, IntegerNullable)
    inferred_type = large_df.ww.logical_types["bool_nullable"]
    assert isinstance(inferred_type, BooleanNullable)
    inferred_type = large_df.ww.logical_types["floats"]
    assert isinstance(inferred_type, Double)
    inferred_type = large_df.ww.logical_types["constant"]
    assert isinstance(inferred_type, Unknown)


def test_url_inference(urls):
    dtypes = ["object", "string"]
    if _is_spark_series(urls[0]):
        dtypes = get_spark_dtypes(dtypes)

    for series in urls:
        for dtype in dtypes:
            inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
            assert isinstance(inferred_type, URL)


def test_phone_inference(phone):
    dtypes = ["object", "string"]
    if _is_spark_series(phone[0]):
        dtypes = get_spark_dtypes(dtypes)

    for series in phone:
        for dtype in dtypes:
            inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
            assert isinstance(inferred_type, PhoneNumber)


def test_ip_inference(ip):
    dtypes = ["object", "string"]
    if _is_spark_series(ip[0]):
        dtypes = get_spark_dtypes(dtypes)

    for series in ip:
        for dtype in dtypes:
            inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
            assert isinstance(inferred_type, IPAddress)



def _get_recent_value_counts(column, num_x):
    """Get the the number of occurrences of the x most recent values in a datetime column.

    Args:
        column (pd.Series): data to use find value counts
        num_x (int): the number of values to retrieve

    Returns:
        value_counts (list(dict)): a list of dictionary with keys `value` and
            `count`. Output is sorted in descending order based on the value counts.
    """
    datetimes = getattr(column.dt, "date")
    frequencies = datetimes.value_counts(dropna=False)
    values = frequencies.sort_index(ascending=False)[:num_x]
    df = values.reset_index()
    df.columns = ["value", "count"]
    df = df.sort_values(["count", "value"], ascending=[False, True])
    value_counts = list(df.to_dict(orient="index").values())
    return value_counts



import pandas as pd

from woodwork.statistics_utils.frequency_inference._constants import (
    ESTIMATED_COLUMN_NAME,
    OBSERVED_COLUMN_NAME,
)
from woodwork.statistics_utils.frequency_inference._get_ranges import _get_ranges
from woodwork.statistics_utils.frequency_inference._types import RangeObject


def _determine_missing_values(estimated_ts, observed_ts):
    """Calculate Missing Values in time_series:

    Args:
        estimated_ts (Series): The estimated time_series.
        observed_ts (Series): The observed time_series.

    Returns:
        (list(RangeObject)): A list of RangeObject data objects. A RangeObject has the following properties:

        - dt: an ISO 8601 formatted string of the first missing timestamp
        - idx: first index of the missing timestamp. Index is relative to estimated timeseries
        - range: the number of sequential elements that are missing
    """
    estimated_df = pd.DataFrame({ESTIMATED_COLUMN_NAME: estimated_ts})
    observed_df = pd.DataFrame({OBSERVED_COLUMN_NAME: observed_ts})

    merged_df = estimated_df.merge(
        observed_df,
        how="left",
        left_on=ESTIMATED_COLUMN_NAME,
        right_on=OBSERVED_COLUMN_NAME,
    )

    observed_null = merged_df[merged_df[OBSERVED_COLUMN_NAME].isnull()]

    if len(observed_null) == 0:
        return []
    ranges = _get_ranges(observed_null.index)

    return [
        RangeObject(
            dt=estimated_ts[start_idx].isoformat(),
            idx=int(start_idx),
            range=int(end_idx - start_idx + 1),
        )
        for start_idx, end_idx in ranges
    ]



import json
import os
import shutil
from pathlib import Path
from unittest.mock import patch

import boto3
import pandas as pd
import pyarrow as pa
import pytest

from woodwork.accessor_utils import _is_dask_dataframe, _is_spark_dataframe
from woodwork.deserialize import from_disk, read_woodwork_table
from woodwork.deserializers.deserializer_base import _check_schema_version
from woodwork.exceptions import (
    OutdatedSchemaWarning,
    ParametersIgnoredWarning,
    UpgradeSchemaWarning,
    WoodworkFileExistsError,
    WoodworkNotInitError,
)
from woodwork.logical_types import Categorical, Ordinal
from woodwork.serializers import get_serializer
from woodwork.serializers.serializer_base import SCHEMA_VERSION, typing_info_to_dict
from woodwork.tests.testing_utils import to_pandas

BUCKET_NAME = "test-bucket"
WRITE_KEY_NAME = "test-key"
TEST_S3_URL = "s3://{}/{}".format(BUCKET_NAME, WRITE_KEY_NAME)
TEST_FILE = "test_serialization_woodwork_table_schema_{}.tar".format(SCHEMA_VERSION)
S3_URL = "s3://woodwork-static/" + TEST_FILE
URL = "https://woodwork-static.s3.amazonaws.com/" + TEST_FILE
TEST_KEY = "test_access_key_es"


def xfail_tmp_disappears(dataframe):
    # TODO: tmp file disappears after deserialize step, cannot check equality with Dask
    if not isinstance(dataframe, pd.DataFrame):
        pytest.xfail(
            "tmp file disappears after deserialize step, cannot check equality with Dask",
        )


def test_error_before_table_init(sample_df, tmpdir):
    error_message = "Woodwork not initialized for this DataFrame. Initialize by calling DataFrame.ww.init"

    with pytest.raises(WoodworkNotInitError, match=error_message):
        sample_df.ww.to_dictionary()

    with pytest.raises(WoodworkNotInitError, match=error_message):
        sample_df.ww.to_disk(str(tmpdir), format="csv")


def test_to_dictionary(sample_df):
    if _is_dask_dataframe(sample_df):
        table_type = "dask"
        age_cat_type_dict = {
            "type": "category",
            "cat_values": [25, 33, 57],
            "cat_dtype": "int64",
        }
        cat_type_dict = {
            "type": "category",
            "cat_values": ["a", "b", "c"],
            "cat_dtype": "object",
        }
    elif _is_spark_dataframe(sample_df):
        table_type = "spark"
        age_cat_type_dict = {"type": "string"}
        cat_type_dict = {"type": "string"}
    else:
        table_type = "pandas"
        age_cat_type_dict = {
            "type": "category",
            "cat_values": [25, 33, 57],
            "cat_dtype": "int64",
        }
        cat_type_dict = {
            "type": "category",
            "cat_values": ["a", "b", "c"],
            "cat_dtype": "object",
        }

    int_val = "int64"
    nullable_int_val = "Int64"
    string_val = "string"
    bool_val = "boolean"
    double_val = "float64"

    expected = {
        "schema_version": SCHEMA_VERSION,
        "name": "test_data",
        "index": "id",
        "time_index": None,
        "column_typing_info": [
            {
                "name": "id",
                "ordinal": 0,
                "use_standard_tags": True,
                "logical_type": {"parameters": {}, "type": "Integer"},
                "physical_type": {"type": int_val},
                "semantic_tags": ["index", "tag1"],
                "description": None,
                "origin": None,
                "metadata": {"is_sorted": True},
            },
            {
                "name": "full_name",
                "ordinal": 1,
                "use_standard_tags": True,
                "logical_type": {"parameters": {}, "type": "Unknown"},
                "physical_type": {"type": string_val},
                "semantic_tags": [],
                "description": None,
                "origin": None,
                "metadata": {},
            },
            {
                "name": "email",
                "ordinal": 2,
                "use_standard_tags": True,
                "logical_type": {"parameters": {}, "type": "EmailAddress"},
                "physical_type": {"type": string_val},
                "semantic_tags": [],
                "description": None,
                "origin": None,
                "metadata": {},
            },
            {
                "name": "phone_number",
                "ordinal": 3,
                "use_standard_tags": True,
                "logical_type": {"parameters": {}, "type": "PhoneNumber"},
                "physical_type": {"type": string_val},
                "semantic_tags": [],
                "description": None,
                "origin": "base",
                "metadata": {},
            },
            {
                "name": "age",
                "ordinal": 4,
                "use_standard_tags": True,
                "logical_type": {
                    "parameters": {"order": [25, 33, 57]},
                    "type": "Ordinal",
                },
                "physical_type": age_cat_type_dict,
                "semantic_tags": ["category"],
                "description": "age of the user",
                "origin": "base",
                "metadata": {"interesting_values": [33, 57]},
            },
            {
                "name": "signup_date",
                "ordinal": 5,
                "use_standard_tags": True,
                "logical_type": {
                    "parameters": {"datetime_format": None, "timezone": None},
                    "type": "Datetime",
                },
                "physical_type": {"type": "datetime64[ns]"},
                "semantic_tags": [],
                "description": "original signup date",
                "origin": "engineered",
                "metadata": {},
            },
            {
                "name": "is_registered",
                "ordinal": 6,
                "use_standard_tags": True,
                "logical_type": {"parameters": {}, "type": "BooleanNullable"},
                "physical_type": {"type": bool_val},
                "semantic_tags": [],
                "description": None,
                "origin": None,
                "metadata": {},
            },
            {
                "name": "double",
                "ordinal": 7,
                "use_standard_tags": True,
                "logical_type": {"parameters": {}, "type": "Double"},
                "physical_type": {"type": double_val},
                "semantic_tags": ["numeric"],
                "description": None,
                "origin": None,
                "metadata": {},
            },
            {
                "name": "double_with_nan",
                "ordinal": 8,
                "use_standard_tags": True,
                "logical_type": {"parameters": {}, "type": "Double"},
                "physical_type": {"type": double_val},
                "semantic_tags": ["numeric"],
                "description": None,
                "origin": None,
                "metadata": {},
            },
            {
                "name": "integer",
                "ordinal": 9,
                "use_standard_tags": True,
                "logical_type": {"parameters": {}, "type": "Integer"},
                "physical_type": {"type": int_val},
                "semantic_tags": ["numeric"],
                "description": None,
                "origin": None,
                "metadata": {},
            },
            {
                "name": "nullable_integer",
                "ordinal": 10,
                "use_standard_tags": True,
                "logical_type": {"parameters": {}, "type": "IntegerNullable"},
                "physical_type": {"type": nullable_int_val},
                "semantic_tags": ["numeric"],
                "description": None,
                "origin": None,
                "metadata": {},
            },
            {
                "name": "boolean",
                "ordinal": 11,
                "use_standard_tags": True,
                "logical_type": {
                    "parameters": {"cast_nulls_as": None},
                    "type": "Boolean",
                },
                "physical_type": {"type": "bool"},
                "semantic_tags": [],
                "description": None,
                "origin": None,
                "metadata": {},
            },
            {
                "name": "categorical",
                "ordinal": 12,
                "use_standard_tags": True,
                "logical_type": {"parameters": {}, "type": "Categorical"},
                "physical_type": cat_type_dict,
                "semantic_tags": ["category"],
                "description": None,
                "origin": None,
                "metadata": {},
            },
            {
                "name": "datetime_with_NaT",
                "ordinal": 13,
                "use_standard_tags": True,
                "logical_type": {
                    "parameters": {"datetime_format": None, "timezone": None},
                    "type": "Datetime",
                },
                "physical_type": {"type": "datetime64[ns]"},
                "semantic_tags": [],
                "description": None,
                "origin": None,
                "metadata": {},
            },
            {
                "name": "url",
                "ordinal": 14,
                "use_standard_tags": True,
                "logical_type": {"parameters": {}, "type": "URL"},
                "physical_type": {"type": string_val},
                "semantic_tags": [],
                "description": None,
                "origin": None,
                "metadata": {},
            },
            {
                "name": "ip_address",
                "ordinal": 15,
                "use_standard_tags": True,
                "logical_type": {"parameters": {}, "type": "IPAddress"},
                "physical_type": {"type": string_val},
                "semantic_tags": [],
                "description": None,
                "origin": None,
                "metadata": {},
            },
        ],
        "loading_info": {"table_type": table_type},
        "table_metadata": {"date_created": "11/16/20"},
    }
    sample_df.ww.init(
        name="test_data",
        index="id",
        semantic_tags={"id": "tag1"},
        logical_types={"age": Ordinal(order=[25, 33, 57])},
        table_metadata={"date_created": "11/16/20"},
        column_descriptions={
            "signup_date": "original signup date",
            "age": "age of the user",
        },
        column_origins={
            "phone_number": "base",
            "age": "base",
            "signup_date": "engineered",
        },
        column_metadata={
            "id": {"is_sorted": True},
            "age": {"interesting_values": [33, 57]},
        },
    )

    description = sample_df.ww.to_dictionary()
    assert description == expected


def test_unserializable_table(sample_df, tmpdir):
    sample_df.ww.init(
        table_metadata={"not_serializable": sample_df["is_registered"].dtype},
    )

    error = "Woodwork table is not json serializable. Check table and column metadata for values that may not be serializable."
    with pytest.raises(TypeError, match=error):
        sample_df.ww.to_disk(
            str(tmpdir),
            format="csv",
            encoding="utf-8",
            engine="python",
        )


def test_serialize_wrong_format(sample_df, tmpdir):
    error = "must be one of the following formats: csv, pickle, parquet"
    with pytest.raises(ValueError, match=error):
        get_serializer(format="test")


def test_to_csv(sample_df, tmpdir):
    if _is_dask_dataframe(sample_df):
        # Dask errors with pd.NA in some partitions, but not others
        sample_df["age"] = sample_df["age"].fillna(25)
    sample_df.ww.init(
        name="test_data",
        index="id",
        semantic_tags={"id": "tag1"},
        logical_types={"age": Ordinal(order=[25, 33, 57])},
        column_descriptions={
            "signup_date": "original signup date",
            "age": "age of the user",
        },
        column_origins={
            "phone_number": "base",
            "age": "base",
            "signup_date": "engineered",
        },
        column_metadata={
            "id": {"is_sorted": True},
            "age": {"interesting_values": [33, 57]},
        },
    )
    sample_df.ww.to_disk(str(tmpdir), format="csv", encoding="utf-8", engine="python")
    deserialized_df = read_woodwork_table(str(tmpdir))

    pd.testing.assert_frame_equal(
        to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True),
        to_pandas(sample_df, index=sample_df.ww.index, sort_index=True),
    )
    assert deserialized_df.ww.schema == sample_df.ww.schema


@pytest.mark.parametrize("format", ["csv", "parquet", "pickle"])
def test_to_disk_with_whitespace(whitespace_df, tmpdir, format):
    df = whitespace_df.copy()
    df.ww.init(index="id", logical_types={"comments": "NaturalLanguage"})
    if format == "pickle" and not isinstance(df, pd.DataFrame):
        msg = "DataFrame type not compatible with pickle serialization. Please serialize to another format."
        with pytest.raises(ValueError, match=msg):
            df.ww.to_disk(str(tmpdir), format="pickle")
    else:
        df.ww.to_disk(str(tmpdir), format=format)
        if format == "parquet":
            filename = "data.parquet"
            format = None
            if _is_dask_dataframe(whitespace_df) or _is_spark_dataframe(whitespace_df):
                filename = None
                format = "parquet"
            deserialized_df = read_woodwork_table(
                path=str(tmpdir),
                filename=filename,
                format=format,
            )
        else:
            deserialized_df = read_woodwork_table(str(tmpdir))
        assert deserialized_df.ww.schema == df.ww.schema
        pd.testing.assert_frame_equal(
            to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True),
            to_pandas(df, index=df.ww.index, sort_index=True),
        )


def test_to_csv_use_standard_tags(sample_df, tmpdir):
    no_standard_tags_df = sample_df.copy()
    no_standard_tags_df.ww.init(use_standard_tags=False)

    no_standard_tags_df.ww.to_disk(
        str(tmpdir),
        format="csv",
        encoding="utf-8",
        engine="python",
    )
    deserialized_no_tags_df = read_woodwork_table(str(tmpdir))
    shutil.rmtree(str(tmpdir))

    standard_tags_df = sample_df.copy()
    standard_tags_df.ww.init(use_standard_tags=True)

    standard_tags_df.ww.to_disk(
        str(tmpdir),
        format="csv",
        encoding="utf-8",
        engine="python",
    )
    deserialized_tags_df = read_woodwork_table(str(tmpdir))
    shutil.rmtree(str(tmpdir))

    assert no_standard_tags_df.ww.schema != standard_tags_df.ww.schema

    assert deserialized_no_tags_df.ww.schema == no_standard_tags_df.ww.schema
    assert deserialized_tags_df.ww.schema == standard_tags_df.ww.schema


def test_deserialize_handles_indexes(sample_df, tmpdir):
    sample_df.ww.init(
        name="test_data",
        index="id",
        time_index="signup_date",
    )
    sample_df.ww.to_disk(str(tmpdir), format="csv")
    deserialized_df = read_woodwork_table(str(tmpdir))
    assert deserialized_df.ww.index == "id"
    assert deserialized_df.ww.time_index == "signup_date"


@pytest.mark.parametrize(
    "file_format",
    ["csv", "pickle", "parquet", "arrow", "feather", "orc"],
)
def test_to_disk(sample_df, tmpdir, file_format):
    if file_format in ("arrow", "feather") and not isinstance(sample_df, pd.DataFrame):
        pytest.xfail("Arrow IPC format (Feather) not supported on Dask or Spark")

    sample_df.ww.init(index="id")
    error_msg = None
    if file_format == "orc" and _is_dask_dataframe(sample_df):
        error_msg = "DataFrame type not compatible with orc serialization. Please serialize to another format."
        error_type = ValueError
    elif file_format == "pickle" and not isinstance(sample_df, pd.DataFrame):
        error_msg = "DataFrame type not compatible with pickle serialization. Please serialize to another format."
        error_type = ValueError

    if error_msg:
        with pytest.raises(error_type, match=error_msg):
            sample_df.ww.to_disk(str(tmpdir), format=file_format)
    else:
        sample_df.ww.to_disk(str(tmpdir), format=file_format)
        if file_format == "parquet":
            filename = "data.parquet"
            format = None
            if _is_dask_dataframe(sample_df) or _is_spark_dataframe(sample_df):
                filename = None
                format = "parquet"
            deserialized_df = read_woodwork_table(
                path=str(tmpdir),
                filename=filename,
                format=format,
            )
        else:
            deserialized_df = read_woodwork_table(str(tmpdir))
        pd.testing.assert_frame_equal(
            to_pandas(sample_df, index=sample_df.ww.index, sort_index=True),
            to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True),
        )
        assert sample_df.ww.schema == deserialized_df.ww.schema


@pytest.mark.parametrize(
    "file_format",
    ["csv", "pickle", "parquet", "arrow", "feather", "orc"],
)
def test_to_disk_custom_data_filename(sample_df, tmpdir, file_format):
    if file_format in ("arrow", "feather") and not isinstance(sample_df, pd.DataFrame):
        pytest.xfail("Arrow IPC format (Feather) not supported on Dask or Spark")

    sample_df.ww.init(index="id")
    error_msg = None
    if file_format == "orc" and _is_dask_dataframe(sample_df):
        error_msg = "DataFrame type not compatible with orc serialization. Please serialize to another format."
        error_type = ValueError
    elif file_format == "pickle" and not isinstance(sample_df, pd.DataFrame):
        error_msg = "DataFrame type not compatible with pickle serialization. Please serialize to another format."
        error_type = ValueError
    elif file_format == "parquet" and _is_dask_dataframe(sample_df):
        error_msg = "Writing a Dask dataframe to parquet with a filename specified is not supported"
        error_type = ValueError
    elif file_format == "csv" and _is_spark_dataframe(sample_df):
        error_msg = "Writing a Spark dataframe to csv with a filename specified is not supported"
        error_type = ValueError
    elif file_format == "parquet" and _is_spark_dataframe(sample_df):
        error_msg = "Writing a Spark dataframe to parquet with a filename specified is not supported"
        error_type = ValueError

    data_filename = f"custom_data.{file_format}"
    filename_to_check = data_filename
    if _is_dask_dataframe(sample_df):
        data_filename = f"custom_data-*.{file_format}"
        filename_to_check = f"custom_data-0.{file_format}"

    if error_msg:
        with pytest.raises(error_type, match=error_msg):
            sample_df.ww.to_disk(
                path=str(tmpdir),
                format=file_format,
                filename=data_filename,
            )
    else:
        sample_df.ww.to_disk(
            path=str(tmpdir),
            format=file_format,
            filename=data_filename,
        )
        assert os.path.isfile(os.path.join(tmpdir, "data", filename_to_check))
        deserialized_df = read_woodwork_table(
            path=str(tmpdir),
            filename=data_filename,
        )
        pd.testing.assert_frame_equal(
            to_pandas(sample_df, index=sample_df.ww.index, sort_index=True),
            to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True),
        )
        assert sample_df.ww.schema == deserialized_df.ww.schema


@pytest.mark.parametrize("file_format", ["csv", "pickle", "arrow", "feather", "orc"])
def test_to_disk_custom_typing_filename(sample_df, tmpdir, file_format):
    if file_format in ("arrow", "feather") and not isinstance(sample_df, pd.DataFrame):
        pytest.xfail("Arrow IPC format (Feather) not supported on Dask or Spark")

    sample_df.ww.init(index="id")
    error_msg = None
    if file_format == "orc" and _is_dask_dataframe(sample_df):
        error_msg = "DataFrame type not compatible with orc serialization. Please serialize to another format."
        error_type = ValueError
    elif file_format == "pickle" and not isinstance(sample_df, pd.DataFrame):
        error_msg = "DataFrame type not compatible with pickle serialization. Please serialize to another format."
        error_type = ValueError

    custom_typing_filename = "custom_typing_info.json"
    if error_msg:
        with pytest.raises(error_type, match=error_msg):
            sample_df.ww.to_disk(
                str(tmpdir),
                format=file_format,
                typing_info_filename=custom_typing_filename,
            )
    else:
        sample_df.ww.to_disk(
            str(tmpdir),
            format=file_format,
            typing_info_filename=custom_typing_filename,
        )
        assert os.path.isfile(os.path.join(tmpdir, custom_typing_filename))
        deserialized_df = read_woodwork_table(
            str(tmpdir),
            typing_info_filename=custom_typing_filename,
        )
        pd.testing.assert_frame_equal(
            to_pandas(sample_df, index=sample_df.ww.index, sort_index=True),
            to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True),
        )
        assert sample_df.ww.schema == deserialized_df.ww.schema


@pytest.mark.parametrize(
    "file_format",
    ["csv", "pickle", "parquet", "arrow", "feather", "orc"],
)
@pytest.mark.parametrize("data_subdirectory", ["custom_data_directory", None])
def test_to_disk_custom_data_subdirectory(
    sample_df,
    tmpdir,
    file_format,
    data_subdirectory,
):
    if file_format in ("arrow", "feather") and not isinstance(sample_df, pd.DataFrame):
        pytest.xfail("Arrow IPC format (Feather) not supported on Dask or Spark")

    sample_df.ww.init(index="id")
    error_msg = None
    if file_format == "orc" and _is_dask_dataframe(sample_df):
        error_msg = "DataFrame type not compatible with orc serialization. Please serialize to another format."
        error_type = ValueError
    elif file_format == "pickle" and not isinstance(sample_df, pd.DataFrame):
        error_msg = "DataFrame type not compatible with pickle serialization. Please serialize to another format."
        error_type = ValueError

    if error_msg:
        with pytest.raises(error_type, match=error_msg):
            sample_df.ww.to_disk(
                str(tmpdir),
                format=file_format,
                data_subdirectory=data_subdirectory,
            )
        shutil.rmtree(str(tmpdir))

    else:
        sample_df.ww.to_disk(
            str(tmpdir),
            format=file_format,
            data_subdirectory=data_subdirectory,
        )
        if data_subdirectory:
            assert os.path.exists(os.path.join(tmpdir, data_subdirectory))
        filename = None
        format = None
        if file_format == "parquet":
            if _is_dask_dataframe(sample_df) or _is_spark_dataframe(sample_df):
                format = "parquet"
            else:
                filename = "data.parquet"

        deserialized_df = read_woodwork_table(
            str(tmpdir),
            filename=filename,
            data_subdirectory=data_subdirectory,
            format=format,
        )
        pd.testing.assert_frame_equal(
            to_pandas(sample_df, index=sample_df.ww.index, sort_index=True),
            to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True),
        )
        assert sample_df.ww.schema == deserialized_df.ww.schema
        shutil.rmtree(str(tmpdir))


@pytest.mark.parametrize(
    "file_format",
    ["csv", "pickle", "parquet", "arrow", "feather", "orc"],
)
def test_to_disk_with_latlong(latlong_df, tmpdir, file_format):
    if file_format in ("arrow", "feather") and not isinstance(latlong_df, pd.DataFrame):
        pytest.xfail("Arrow IPC format (Feather) not supported on Dask or Spark")
    latlong_df.ww.init(logical_types={col: "LatLong" for col in latlong_df.columns})

    error_msg = None
    if file_format == "orc" and _is_dask_dataframe(latlong_df):
        error_msg = "DataFrame type not compatible with orc serialization. Please serialize to another format."
        error_type = ValueError
    elif file_format == "pickle" and not isinstance(latlong_df, pd.DataFrame):
        error_msg = "DataFrame type not compatible with pickle serialization. Please serialize to another format."
        error_type = ValueError

    if error_msg:
        with pytest.raises(error_type, match=error_msg):
            latlong_df.ww.to_disk(str(tmpdir), format=file_format)
    else:
        latlong_df.ww.to_disk(str(tmpdir), format=file_format)
        filename = None
        format = None
        if file_format == "parquet":
            if _is_dask_dataframe(latlong_df) or _is_spark_dataframe(latlong_df):
                format = "parquet"
            else:
                filename = "data.parquet"

        deserialized_df = read_woodwork_table(
            str(tmpdir),
            filename=filename,
            format=format,
        )

        pd.testing.assert_frame_equal(
            to_pandas(latlong_df).reset_index(drop=True),
            to_pandas(deserialized_df).reset_index(drop=True),
        )
        assert latlong_df.ww.schema == deserialized_df.ww.schema


def test_to_disk_parquet_no_file_extension(sample_df, tmpdir):
    if _is_dask_dataframe(sample_df) or _is_spark_dataframe(sample_df):
        pytest.skip(
            "Specifying filename for writing Dask or Spark DataFrames to parquet is not supported.",
        )
    sample_df.ww.init(index="id")
    sample_df.ww.to_disk(str(tmpdir), filename="parquet_data", format="parquet")

    error_msg = "Could not determine format. Please specify filename and/or format."
    # Without specifying format, WW doens't know what type of file this is
    with pytest.raises(ValueError, match=error_msg):
        deserialized_df = read_woodwork_table(
            str(tmpdir),
            filename="parquet_data",
        )

    deserialized_df = read_woodwork_table(
        str(tmpdir),
        filename="parquet_data",
        format="parquet",
    )
    pd.testing.assert_frame_equal(
        to_pandas(sample_df, index=sample_df.ww.index, sort_index=True),
        to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True),
    )


def test_to_disk_parquet_typing_info_file_is_none(sample_df, tmpdir):
    sample_df.ww.init(index="id")
    sample_df.ww.to_disk(str(tmpdir), format="parquet")

    filename = "data.parquet"
    format = None
    if _is_dask_dataframe(sample_df) or _is_spark_dataframe(sample_df):
        filename = None
        format = "parquet"
    deserialized_df = read_woodwork_table(
        str(tmpdir),
        filename=filename,
        typing_info_filename=None,
        format=format,
    )
    pd.testing.assert_frame_equal(
        to_pandas(sample_df, index=sample_df.ww.index, sort_index=True),
        to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True),
    )


def test_to_disk_parquet_warns_if_typing_info_file_specified(sample_df, tmpdir):
    sample_df.ww.init()

    message = "Typing info filename has been ignored. Typing information will be stored in parquet file header."
    with pytest.warns(ParametersIgnoredWarning, match=message):
        sample_df.ww.to_disk(
            str(tmpdir),
            format="parquet",
            typing_info_filename="woodwork.json",
        )


def test_to_disk_parquet_saves_custom_metadata_as_expected(sample_df, tmpdir):
    sample_df.ww.init(index="id")
    sample_df.ww.set_types(
        logical_types={"categorical": "CountryCode"},
        semantic_tags={"age": "age"},
    )
    sample_df.ww.to_disk(str(tmpdir), format="parquet")
    expected_typing_info = typing_info_to_dict(sample_df)

    if _is_dask_dataframe(sample_df):
        filename = "part.0.parquet"
        path = os.path.join(tmpdir, "data", filename)
    elif _is_spark_dataframe(sample_df):
        path = os.path.join(tmpdir, "data")
        files = os.listdir(path)
        metadata_file = sorted([f for f in files if Path(f).suffix == ".parquet"])[0]
        path = os.path.join(path, metadata_file)
    else:
        filename = "data.parquet"
        path = os.path.join(tmpdir, "data", filename)

    file_metadata = pa.parquet.read_metadata(path).metadata
    assert b"ww_meta" in file_metadata.keys()
    ww_meta = json.loads(file_metadata[b"ww_meta"])
    columns = ww_meta["column_typing_info"]
    cat_info = list(filter(lambda col: col["name"] == "categorical", columns))[0]
    age_info = list(filter(lambda col: col["name"] == "age", columns))[0]

    assert cat_info["logical_type"]["type"] == "CountryCode"
    assert "age" in age_info["semantic_tags"]

    # location, type and params are added during serialization, so they are not present
    # in the expected typing information created from the Woodwork dataframe.
    del ww_meta["loading_info"]["location"]
    del ww_meta["loading_info"]["type"]
    del ww_meta["loading_info"]["params"]

    assert ww_meta == expected_typing_info


def test_categorical_dtype_serialization(serialize_df, tmpdir):
    ltypes = {
        "cat_int": Categorical,
        "ord_int": Ordinal(order=[1, 2]),
        "cat_float": Categorical,
        "ord_float": Ordinal(order=[1.0, 2.0]),
        "cat_bool": Categorical,
        "ord_bool": Ordinal(order=[True, False]),
    }
    if isinstance(serialize_df, pd.DataFrame):
        formats = ["csv", "pickle", "parquet"]
    else:
        formats = ["csv"]

    for format in formats:
        df = serialize_df.copy()
        df.ww.init(index="id", logical_types=ltypes)
        df.ww.to_disk(str(tmpdir), format=format)
        deserialized_df = read_woodwork_table(str(tmpdir), filename=f"data.{format}")
        pd.testing.assert_frame_equal(
            to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True),
            to_pandas(df, index=df.ww.index, sort_index=True),
        )
        assert deserialized_df.ww.schema == df.ww.schema
        shutil.rmtree(str(tmpdir))


@pytest.fixture
def s3_client():
    # TODO: Fix Moto tests needing to explicitly set permissions for objects
    _environ = os.environ.copy()
    from moto import mock_s3

    with mock_s3():
        s3 = boto3.resource("s3")
        yield s3
    os.environ.clear()
    os.environ.update(_environ)


@pytest.fixture
def s3_bucket(s3_client):
    s3_client.create_bucket(Bucket=BUCKET_NAME, ACL="public-read-write")
    s3_bucket = s3_client.Bucket(BUCKET_NAME)
    yield s3_bucket


def make_public(s3_client, s3_bucket):
    obj = list(s3_bucket.objects.all())[0].key
    s3_client.ObjectAcl(BUCKET_NAME, obj).put(ACL="public-read-write")


@pytest.mark.parametrize("profile_name", [None, False])
def test_to_csv_S3(sample_df, s3_client, s3_bucket, profile_name):
    xfail_tmp_disappears(sample_df)

    sample_df.ww.init(
        name="test_data",
        index="id",
        semantic_tags={"id": "tag1"},
        logical_types={"age": Ordinal(order=[25, 33, 57])},
    )
    sample_df.ww.to_disk(
        TEST_S3_URL,
        format="csv",
        encoding="utf-8",
        engine="python",
        profile_name=profile_name,
    )
    make_public(s3_client, s3_bucket)

    deserialized_df = read_woodwork_table(TEST_S3_URL, profile_name=profile_name)

    pd.testing.assert_frame_equal(
        to_pandas(sample_df, index=sample_df.ww.index, sort_index=True),
        to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True),
    )
    assert sample_df.ww.schema == deserialized_df.ww.schema


@pytest.mark.parametrize("profile_name", [None, False])
def test_serialize_s3_pickle(sample_df_pandas, s3_client, s3_bucket, profile_name):
    sample_df_pandas.ww.init()
    sample_df_pandas.ww.to_disk(TEST_S3_URL, format="pickle", profile_name=profile_name)
    make_public(s3_client, s3_bucket)
    deserialized_df = read_woodwork_table(TEST_S3_URL, profile_name=profile_name)

    pd.testing.assert_frame_equal(
        to_pandas(sample_df_pandas, index=sample_df_pandas.ww.index, sort_index=True),
        to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True),
    )
    assert sample_df_pandas.ww.schema == deserialized_df.ww.schema


@pytest.mark.parametrize("profile_name", [None, False])
def test_serialize_s3_parquet(sample_df, s3_client, s3_bucket, profile_name):
    xfail_tmp_disappears(sample_df)

    sample_df.ww.init()
    sample_df.ww.to_disk(TEST_S3_URL, format="parquet", profile_name=profile_name)
    make_public(s3_client, s3_bucket)
    deserialized_df = read_woodwork_table(
        TEST_S3_URL,
        filename="data.parquet",
        profile_name=profile_name,
    )

    pd.testing.assert_frame_equal(
        to_pandas(sample_df, index=sample_df.ww.index, sort_index=True),
        to_pandas(deserialized_df, index=deserialized_df.ww.index, sort_index=True),
    )
    assert sample_df.ww.schema == deserialized_df.ww.schema


def create_test_credentials(test_path):
    with open(test_path, "w+") as f:
        f.write("[test]\n")
        f.write("aws_access_key_id=AKIAIOSFODNN7EXAMPLE\n")
        f.write("aws_secret_access_key=wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY\n")


def create_test_config(test_path_config):
    with open(test_path_config, "w+") as f:
        f.write("[profile test]\n")
        f.write("region=us-east-2\n")
        f.write("output=text\n")


@pytest.fixture
def setup_test_profile(monkeypatch, tmpdir):
    cache = str(tmpdir.join(".cache").mkdir())
    test_path = os.path.join(cache, "test_credentials")
    test_path_config = os.path.join(cache, "test_config")
    monkeypatch.setenv("AWS_SHARED_CREDENTIALS_FILE", test_path)
    monkeypatch.setenv("AWS_CONFIG_FILE", test_path_config)
    monkeypatch.delenv("AWS_ACCESS_KEY_ID", raising=False)
    monkeypatch.delenv("AWS_SECRET_ACCESS_KEY", raising=False)
    monkeypatch.setenv("AWS_PROFILE", "test")

    try:
        os.remove(test_path)
    except OSError:
        pass
    try:
        os.remove(test_path_config)
    except OSError:
        pass

    create_test_credentials(test_path)
    create_test_config(test_path_config)
    yield
    os.remove(test_path)
    os.remove(test_path_config)


def test_s3_test_profile(sample_df, s3_client, s3_bucket, setup_test_profile):
    xfail_tmp_disappears(sample_df)
    sample_df.ww.init()
    sample_df.ww.to_disk(
        TEST_S3_URL,
        format="csv",
        encoding="utf-8",
        engine="python",
        profile_name="test",
    )
    make_public(s3_client, s3_bucket)
    deserialized_df = read_woodwork_table(TEST_S3_URL, profile_name="test")

    pd.testing.assert_frame_equal(
        to_pandas(sample_df, index=sample_df.ww.index),
        to_pandas(deserialized_df, index=deserialized_df.ww.index),
    )
    assert sample_df.ww.schema == deserialized_df.ww.schema


def test_serialize_url_csv(sample_df):
    sample_df.ww.init()
    error_text = "Writing to URLs is not supported"
    with pytest.raises(ValueError, match=error_text):
        sample_df.ww.to_disk(URL, format="csv", encoding="utf-8", engine="python")


def test_serialize_subdirs_not_removed(sample_df, tmpdir):
    sample_df.ww.init()
    write_path = tmpdir.mkdir("test")
    test_dir = write_path.mkdir("test_dir")
    sample_text = str(test_dir.join("sample_text.json"))

    with open(sample_text, "w") as f:
        json.dump("__SAMPLE_TEXT__", f)

    sample_df.ww.to_disk(
        path=str(write_path),
        index="1",
        sep="\t",
        compression=None,
        encoding="utf-8",
        typing_info_filename="woodwork_typing_info_2.json",
    )

    assert os.path.exists(sample_text)
    with open(sample_text, "r") as f:
        assert "__SAMPLE_TEXT__" in json.load(f)

    shutil.rmtree(str(tmpdir))


@pytest.mark.parametrize("profile_name", [None, False])
def test_deserialize_url_csv(sample_df_pandas, profile_name):
    sample_df_pandas.ww.init(index="id")
    deserialized_df = read_woodwork_table(URL, profile_name=profile_name)
    pd.testing.assert_frame_equal(
        to_pandas(sample_df_pandas, index=sample_df_pandas.ww.index),
        to_pandas(deserialized_df, index=deserialized_df.ww.index),
    )
    assert sample_df_pandas.ww.schema == deserialized_df.ww.schema


def test_deserialize_s3_csv(sample_df_pandas):
    sample_df_pandas.ww.init(index="id")
    deserialized_df = read_woodwork_table(S3_URL, profile_name=False)

    pd.testing.assert_frame_equal(
        to_pandas(sample_df_pandas, index=sample_df_pandas.ww.index),
        to_pandas(deserialized_df, index=deserialized_df.ww.index),
    )
    assert sample_df_pandas.ww.schema == deserialized_df.ww.schema


@patch("woodwork.table_accessor._validate_accessor_params")
def test_deserialize_validation_control(mock_validate_accessor_params):
    assert not mock_validate_accessor_params.called
    read_woodwork_table(URL)
    assert not mock_validate_accessor_params.called
    read_woodwork_table(URL, validate=True)
    assert mock_validate_accessor_params.called


def test_check_later_schema_version():
    def test_version(major, minor, patch, raises=True):
        version_to_check = ".".join([str(v) for v in [major, minor, patch]])
        if raises:
            warning_text = (
                "The schema version of the saved Woodwork table "
                "%s is greater than the latest supported %s. "
                "You may need to upgrade woodwork. Attempting to load Woodwork table ..."
                % (version_to_check, SCHEMA_VERSION)
            )
            with pytest.warns(UpgradeSchemaWarning, match=warning_text):
                _check_schema_version(version_to_check)
        else:
            with pytest.warns(None) as record:
                _check_schema_version(version_to_check)
            assert len(record) == 0

    major, minor, patch = [int(s) for s in SCHEMA_VERSION.split(".")]

    test_version(major + 1, minor, patch)
    test_version(major, minor + 1, patch)
    test_version(major, minor, patch + 1)
    test_version(major, minor - 1, patch + 1, raises=False)


def test_earlier_schema_version():
    def test_version(major, minor, patch, raises=True):
        version_to_check = ".".join([str(v) for v in [major, minor, patch]])
        if raises:
            warning_text = (
                "The schema version of the saved Woodwork table "
                "%s is no longer supported by this version "
                "of woodwork. Attempting to load Woodwork table ..."
                % (version_to_check)
            )
            with pytest.warns(OutdatedSchemaWarning, match=warning_text):
                _check_schema_version(version_to_check)
        else:
            with pytest.warns(None) as record:
                _check_schema_version(version_to_check)
            assert len(record) == 0

    major, minor, patch = [int(s) for s in SCHEMA_VERSION.split(".")]

    test_version(major - 1, minor, patch)
    test_version(major, minor - 1, patch, raises=False)
    test_version(major, minor, patch - 1, raises=False)


@pytest.mark.parametrize("format", ["csv", "parquet", "pickle"])
def test_overwrite_error(sample_df, tmpdir, format):
    if format == "pickle" and (
        _is_dask_dataframe(sample_df) or _is_spark_dataframe(sample_df)
    ):
        pytest.skip("Cannot pickle dask and spark dataframes")

    folder_1 = str(tmpdir.join("folder_1"))
    folder_2 = str(tmpdir.join("folder_2"))
    sample_df.ww.init()

    if format != "parquet":
        # Parquet does not use typing info file
        sample_df.ww.to_disk(folder_1, data_subdirectory=None, format=format)
        with pytest.raises(WoodworkFileExistsError, match="Typing info already exists"):
            sample_df.ww.to_disk(folder_1, format=format)

    sample_df.ww.to_disk(folder_2, data_subdirectory=None, format=format)
    with pytest.raises(WoodworkFileExistsError, match="Data file already exists"):
        sample_df.ww.to_disk(
            folder_2,
            format=format,
            typing_info_filename="new_typing_info",
            data_subdirectory=None,
        )

    shutil.rmtree(str(tmpdir))


@patch("woodwork.deserialize.read_woodwork_table")
def test_from_disk(mock_read_woodwork_table, tmpdir, sample_df):
    sample_df.ww.init(
        name="test_data",
        index="id",
        time_index="signup_date",
    )
    sample_df.ww.to_disk(str(tmpdir), format="csv")

    expected_params = {
        "filename": "some_name",
        "data_subdirectory": "data",
        "typing_info_filename": "woodwork_typing_info_random.json",
        "profile_name": None,
        "validate": True,
    }

    _ = from_disk(str(tmpdir), **expected_params)
    mock_read_woodwork_table.assert_called_once_with(str(tmpdir), **expected_params)



from woodwork.deserializers.utils import _get_deserializer


def from_disk(
    path,
    filename=None,
    data_subdirectory="data",
    typing_info_filename="woodwork_typing_info.json",
    profile_name=None,
    validate=False,
    **kwargs,
):
    """Convenience function to call `read_woodwork_table`."""
    deserialized_df = read_woodwork_table(
        path,
        filename=filename,
        data_subdirectory=data_subdirectory,
        typing_info_filename=typing_info_filename,
        profile_name=profile_name,
        validate=validate,
        **kwargs,
    )

    return deserialized_df


def read_woodwork_table(
    path,
    filename=None,
    data_subdirectory="data",
    typing_info_filename="woodwork_typing_info.json",
    profile_name=None,
    validate=False,
    format=None,
    **kwargs,
):
    """Read Woodwork table from disk, S3 path, or URL.

    Args:
        path (str): Directory on disk, S3 path, or URL to read data and typing information.
        filename (str, optional): The name of the file used to store the data during serialization. If not specified, will be
            determined from the typing info file. Must be specified when deserializing from a single parquet file.
        data_subdirectory (str, optional): The subdirectory in which the data was stored during serialization. Defaults to "data".
        typing_info_filename (str, optional): The name of the JSON file used to store the Woodwork typing information during
            serialization. Defaults to "woodwork_typing_info.json".
        format (str, optional): The format used to serialize the data. Required if the serialized filename suffix does not
            match the format or when deserializing from parquet files into Dask or Spark dataframes.
        profile_name (str, bool): The AWS profile specified to write to S3. Will default to None and search for AWS credentials.
            Set to False to use an anonymous profile.
        validate (bool, optional): Whether parameter and data validation should occur when initializing Woodwork dataframe
            during deserialization. Defaults to False. Note: If serialized data was modified outside of Woodwork and you
            are unsure of the validity of the data or typing information, `validate` should be set to True.
        kwargs (keywords): Additional keyword arguments to pass as keyword arguments to the underlying deserialization method.

    Returns:
        DataFrame: DataFrame with Woodwork typing information initialized.
    """
    deserializer = _get_deserializer(
        path,
        filename,
        data_subdirectory,
        typing_info_filename,
        profile_name,
        format,
    )
    return deserializer.deserialize(profile_name, validate, **kwargs)
